/* hash.c - IRC network state database
 * Copyright 2000-2002 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: hash.c,v 1.159 2002/06/18 23:14:57 zoot Exp $
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#include "common.h"
#include "commands.h"
#include "dict.h"
#include "conf.h"
#include "hash.h"
#include "global.h"
#include "log.h"
#include "nickserv.h"
#include "chanserv.h"
#include "policer.h"
#include "parse.h"
#include "recdb.h"

extern time_t boot_time;
extern struct policer_params *oper_policer_params, *luser_policer_params;

struct server *self;
struct server *servers_num[64*64];
dict_t channels;
dict_t clients;
dict_t servers;
unsigned int max_clients, invis_clients;
time_t max_clients_time;
unsigned int server_count;
struct userList curr_opers;
unsigned int nicklen = NICKLEN, force_n2k;

static unsigned int next_numeric = 0;

static void hash_cleanup(void);

void init_structs(void)
{
	unsigned int i;
	char numer[COMBO_NUMERIC_LEN+1];

	for (i=0; i<ArrayLength(servers_num); i++) {
	    servers_num[i] = NULL;
	}

        channels = dict_new();
        clients = dict_new();
        servers = dict_new();
        userList_init(&curr_opers);
        reg_exit_func(hash_cleanup);

        /* Use a three-digit numeric if we can, else use five. */
        if ((conf_server_num() < 64) && !force_n2k) {
            inttobase64(numer, conf_server_num(), 1);
            numer[1] = ']';
            numer[2] = ']';
            numer[3] = 0;
        } else {
            inttobase64(numer, conf_server_num(), 2);
            numer[2] = ']';
            numer[3] = ']';
            numer[4] = ']';
            numer[5] = 0;
        }
	self = AddServer(NULL, conf_server_name(), 0, boot_time, now, numer, conf_server_desc());
}

void
send_burst(void)
{
	unsigned int i, hop, max_hop=1;
        dict_iterator_t it;
	
	/* burst servers, closest first (but self is sent during auth) */
	for (i=0; i<ArrayLength(servers_num); i++) {
	    if (servers_num[i] && servers_num[i]->hops > max_hop) {
		max_hop = servers_num[i]->hops;
	    }
	}
	for (hop=1; hop<=max_hop; hop++) {
	    for (i=0;i<ArrayLength(servers_num);i++) {
		if (!servers_num[i]) continue;

		if ((servers_num[i]->hops == hop) && (servers_num[i] != self->uplink))
		    irc_server(servers_num[i]);
	    }
	}

	/* burst local nicks */
	for (i=0; i<=self->num_mask; i++) {
	    if (self->users[i]) irc_user(self->users[i]);
	}

	/* burst channels */
        for (it=dict_first(channels); it; it=iter_next(it)) {
            irc_burst(iter_data(it));
        }

	irc_eob();
}

static server_link_func_t *slf_list;
static unsigned int slf_size = 0, slf_used = 0;

void
reg_server_link_func(server_link_func_t handler)
{
    if (slf_used == slf_size) {
	if (slf_size) {
	    slf_size <<= 1;
	    slf_list = realloc(slf_list, slf_size*sizeof(server_link_func_t));
	} else {
	    slf_size = 8;
	    slf_list = malloc(slf_size*sizeof(server_link_func_t));
	}
    }
    slf_list[slf_used++] = handler;
}

struct server *
AddServer(struct server *uplink, const char *name, int hops, time_t boot, time_t link, const char *numeric, const char *description)
{
	struct server* sNode;
        int slen, mlen;

	if ((sNode = GetServer(numeric))) {
            /* This means we're trying to re-add an existant server.
             * To be safe, we should forget the previous incarnation.
             * (And all its linked servers.)
             *
             * It usually only happens in replays when the original
             * had a ping timeout and the replay didn't (because
             * replaying a ping timeout invariably gets things wrong).
             */
            DelServer(sNode, 0, NULL);
        }

        switch (strlen(numeric)) {
        case 5: slen = 2; mlen = 3; break;
        case 4: slen = 1; mlen = 3; break;
        case 3: slen = 1; mlen = 2; break;
        default:
            log(MAIN_LOG, LOG_ERROR, "AddServer(\"%s\", \"%s\", ...): Numeric %s has invalid length.\n", uplink->name, name, numeric);
            return NULL;
        }

        sNode = calloc(1, sizeof(*sNode));
	sNode->uplink = uplink;
	safestrncpy(sNode->name, name, sizeof(sNode->name));
        sNode->num_mask = base64toint(numeric+slen, mlen);
	sNode->hops = hops;
	sNode->clients = 0;
	sNode->boot = boot;
	sNode->link = link;
        strncpy(sNode->numeric, numeric, slen);
	safestrncpy(sNode->description, description, sizeof(sNode->description));
        sNode->users = calloc(sNode->num_mask+1, sizeof(*sNode->users));
        serverList_init(&sNode->children);
        if (sNode->uplink) {
            /* uplink may be NULL if we're just building ourself */
            serverList_append(&sNode->uplink->children, sNode);
        }
        servers_num[base64toint(numeric, slen)] = sNode;
	dict_insert(servers, sNode->name, sNode);

	if (hops && !self->burst) {
	    unsigned int n;
	    for (n=0; n<slf_used; n++) {
		slf_list[n](sNode);
	    }
	}

	server_count++;
	return sNode;
}

void DelServer(struct server* serv, int announce, const char *message)
{
	unsigned int i;

	/* If we receive an ERROR command before the SERVER
	 * command a NULL server can be passed */
	if (!serv) return;

        /* Hrm, what's the right way to SQUIT some other server?
         * (This code is only to handle killing juped servers.) */
	if (announce && (serv->uplink == self) && (serv != self->uplink)) {
            irc_squit(serv, message);
        }

	/* must recursively remove servers linked to this one first */
	for (i=serv->children.used;i>0;) {
            if (serv->children.list[--i] != self) {
                DelServer(serv->children.list[i], false, NULL);
            }
	}

	/* clean up server's user hash tables */
	for (i=0;i<=serv->num_mask;i++) {
	    if (serv->users[i]) {
		DelUser(serv->users[i], NULL, false, "server delinked");
	    }
	}

	/* delete server */
        if (serv->uplink) serverList_remove(&serv->uplink->children, serv);
        if (serv == self->uplink) self->uplink = NULL;
	servers_num[base64toint(serv->numeric, strlen(serv->numeric))] = NULL;
        dict_remove(servers, serv->name);
        serverList_clean(&serv->children);
        free(serv->users);
	free(serv);
	server_count--;
}

/* Numerics can be XYY, XYYY, or XXYYY; with X's identifying the
 * server and Y's indentifying the client on that server. */
struct server*
GetServer(const char *numeric)
{
    switch (strlen(numeric)) {
    default:
        return servers_num[base64toint(numeric, 2)];
    case 4:
    case 3:
    case 1:
        return servers_num[base64toint(numeric, 1)];
    case 0:
        return NULL;
    }
}

struct server*
GetServerH(const char *name)
{
    unsigned int n;
    for (n=0; n<ArrayLength(servers_num); n++) {
	if (!servers_num[n]) continue;
	if (!strcasecmp(name, servers_num[n]->name)) return servers_num[n];
    }
    return NULL;
}

static int
get_local_numeric(void)
{
    if (self->clients >= self->num_mask) return -1;
    while (self->users[next_numeric]) {
        if (++next_numeric > self->num_mask) next_numeric = 0;
    }
    return next_numeric;
}

static void
make_numeric(struct server *svr, int local_num, char *outbuf)
{
    int slen, llen;

    if (force_n2k || svr->numeric[1]) {
        slen = 2;
        llen = 3;
    } else {
        slen = 1;
        llen = (local_num < 64*64) ? 2 : 3;
    }
    strncpy(outbuf, svr->numeric, slen);
    inttobase64(outbuf+slen, local_num, llen);
    outbuf[slen+llen] = 0;
}

struct userNode *
AddService(const char *nick, const char *desc)
{
    char numeric[COMBO_NUMERIC_LEN+1];
    int local_num = get_local_numeric();

    if (local_num == -1) return 0;
    make_numeric(self, local_num, numeric);
    return AddUser(self, nick, nick, self->name, "+oik", numeric, desc, now, "AAAAAA");
}

struct userNode *
AddClone(const char *nick, const char *ident, const char *hostname, const char *desc)
{
    char numeric[COMBO_NUMERIC_LEN+1];
    int local_num = get_local_numeric();

    if (local_num == -1) return 0;
    make_numeric(self, local_num, numeric);
    return AddUser(self, nick, ident, hostname, "+i", numeric, desc, now, "AAAAAA");
}

static new_user_func_t *nuf_list;
static unsigned int nuf_size = 0, nuf_used = 0;

void
reg_new_user_func(new_user_func_t handler)
{
    if (nuf_used == nuf_size) {
	if (nuf_size) {
	    nuf_size <<= 1;
	    nuf_list = realloc(nuf_list, nuf_size*sizeof(new_user_func_t));
	} else {
	    nuf_size = 8;
	    nuf_list = malloc(nuf_size*sizeof(new_user_func_t));
	}
    }
    nuf_list[nuf_used++] = handler;
}

static nick_change_func_t *ncf2_list;
static unsigned int ncf2_size = 0, ncf2_used = 0;

void
reg_nick_change_func(nick_change_func_t handler)
{
    if (ncf2_used == ncf2_size) {
        if (ncf2_size) {
            ncf2_size <<= 1;
            ncf2_list = realloc(ncf2_list, ncf2_size*sizeof(nick_change_func_t));
        } else {
            ncf2_size = 8;
            ncf2_list = malloc(ncf2_size*sizeof(nick_change_func_t));
        }
    }
    ncf2_list[ncf2_used++] = handler;
}


static del_user_func_t *duf_list;
static unsigned int duf_size = 0, duf_used = 0;

void
reg_del_user_func(del_user_func_t handler)
{
    if (duf_used == duf_size) {
	if (duf_size) {
	    duf_size <<= 1;
	    duf_list = realloc(duf_list, duf_size*sizeof(del_user_func_t));
	} else {
	    duf_size = 8;
	    duf_list = malloc(duf_size*sizeof(del_user_func_t));
	}
    }
    duf_list[duf_used++] = handler;
}

void
unreg_del_user_func(del_user_func_t handler)
{
    unsigned int i;
    for (i=0; i<duf_used; i++) {
        if (duf_list[i] == handler) break;
    }
    if (i == duf_used) return;
    memmove(duf_list+i, duf_list+i+1, (duf_used-i-1)*sizeof(duf_list[0]));
    duf_used--;
}

struct userNode*
AddUser(struct server* uplink, const char *nick, const char *ident, const char *hostname, const char *modes, const char *numeric, const char *userinfo, time_t timestamp, const char *realip)
{
	struct userNode *oldUser, *uNode;
	struct policer_params *pol_params;
	unsigned int n;

        if ((strlen(numeric) < 3) || (strlen(numeric) > 5)) {
            log(MAIN_LOG, LOG_WARNING, "AddUser(%p, %s, ...): numeric %s wrong length!\n", uplink, nick, numeric);
            return NULL;
        }

        if (!uplink) {
            log(MAIN_LOG, LOG_WARNING, "AddUser(%p, %s, ...): server for numeric %s doesn't exist!\n", uplink, nick, numeric);
            return NULL;
        }

        if (uplink != GetServer(numeric)) {
            log(MAIN_LOG, LOG_WARNING, "AddUser(%p, %s, ...): server for numeric %s differs from nominal uplink %s.\n", uplink, nick, numeric, uplink->name);
            return NULL;
        }

        if (strlen(nick) > nicklen) {
            log(MAIN_LOG, LOG_ERROR, "AddUser(%p, %s, ...): nick too long (detected max is %d)\n", uplink, nick, nicklen);
            irc_squit(self, "Nick too long; see main.log for details");
            return NULL;
        }

	/* create new usernode and set all values */
	uNode = calloc(1, sizeof(*uNode));
	safestrncpy(uNode->nick, nick, sizeof(uNode->nick));
	safestrncpy(uNode->ident, ident, sizeof(uNode->ident));
	safestrncpy(uNode->info, userinfo, sizeof(uNode->info));
	safestrncpy(uNode->hostname, hostname, sizeof(uNode->hostname));
	safestrncpy(uNode->numeric, numeric, sizeof(uNode->numeric));
	uNode->ip = base64toint(realip, 6);
	uNode->timestamp=timestamp;
	modeList_init(&uNode->channels);
	uNode->uplink = uplink;
        if (++uNode->uplink->clients > uNode->uplink->max_clients) {
            uNode->uplink->max_clients = uNode->uplink->clients;
        }
	pol_params = IsOper(uNode) ? oper_policer_params : luser_policer_params;
	uNode->command_policer = policer_new(pol_params, uNode->timestamp);
	mod_usermode(uNode, modes);
        uNode->num_local = base64toint(numeric+strlen(uNode->uplink->numeric), 3) & uNode->uplink->num_mask;

	/* does the nick already exist?  if so, how to handle? */
	if ((oldUser = GetUserH(nick))) {
	    if (oldUser->uplink == self) {
		/* Since we want services to be authoritative over
		 * other servers, we'll issue a KILL regardless of
		 * timestamp comparisons.  If this other user is +k,
		 * then somebody screwed up.  */
                irc_kill(NULL, uNode, "Nick collision");
	    }
	} else {
            dict_insert(clients, uNode->nick, uNode);
            if (dict_size(clients) > max_clients) {
                max_clients = dict_size(clients);
                max_clients_time = now;
            }
        }

	uNode->uplink->users[uNode->num_local] = uNode;

	if (IsLocal(uNode)) irc_user(uNode);

        for (n=0; n<nuf_used; n++) {
            if (nuf_list[n](uNode)) break;
        }

	return uNode;
}

/* removes user from it's server's hash table and nick hash table */
void
DelUser(struct userNode* user, struct userNode *killer, int announce, const char *why)
{
	unsigned int n;

	/* remove user from all channels */
	for (n=user->channels.used; n>0;) {
	    DelChannelUser(user, user->channels.list[--n]->channel, false, 0);
	}
	/* Call these in reverse order so ChanServ can update presence
	   information before NickServ nukes the handle_info. */
	for (n = duf_used; n > 0; ) duf_list[--n](user);

	user->uplink->clients--;
	user->uplink->users[user->num_local] = NULL;
	if (IsOper(user)) userList_remove(&curr_opers, user);
	/* remove from global dictionary, but only if not after a collide */
        if (user == dict_find(clients, user->nick, NULL)) {
            dict_remove(clients, user->nick);
        }

        if (IsInvisible(user)) invis_clients--;

	if (announce) {
	    if (IsLocal(user)) {
		irc_quit(user, why);
	    } else {
		irc_kill(killer, user, why);
	    }
	}

        modeList_clean(&user->channels);
        policer_delete(user->command_policer);
        /* We don't free them, in case we try to privmsg them or something
         * (like when a stupid oper kills themself).  We just put them onto
         * a list of clients that get freed after processing each line.
         */
	user->dead = 1;
        userList_append(&dead_users, user);
}

/* reintroduces a user after it has been killed. */
void
ReintroduceUser(struct userNode *user)
{	
    unsigned int n;
	
    irc_user(user);

    /* rejoin all channels. */
    for (n = 0; n < user->channels.used; n++) {
        struct modeNode *mn = user->channels.list[n];
        char modes[16];
        switch (mn->modes & (MODE_CHANOP | MODE_VOICE)) {
        case 0: modes[0] = 0; break;
        case MODE_VOICE: snprintf(modes, sizeof(modes), "+v %s", user->numeric); break;
        case MODE_CHANOP: snprintf(modes, sizeof(modes), "+o %s", user->numeric); break;
        case MODE_CHANOP|MODE_VOICE: snprintf(modes, sizeof(modes), "+ov %s %s", user->numeric, user->numeric); break;
        }
	irc_join(user, mn->channel);
        mn->modes &= ~(MODE_CHANOP|MODE_VOICE); /* since they just rejoined */
        if (modes[0]) irc_mode(user, mn->channel, modes);
    }
}

void
NickChange(struct userNode* user, const char *new_nick, int no_announce)
{
    unsigned int nn;

    /* don't do anything if there's no change */
    if (!strncmp(new_nick, user->nick, NICKLEN)) return;

    /* remove old entry from clients dictionary */
    dict_remove(clients, user->nick);

    /* Make callbacks for nick changes.
     * This needs to be done with the old nick still in place, else
     * dict lookups based on user->nick will miss the user.
     * It is safe to re-insert using new_nick (which may be a
     * temporary pointer) since the later value of user->nick will be
     * correct. */
    for (nn=0; nn<ncf2_used; nn++) {
        ncf2_list[nn](user, new_nick);
    }

    /* copy new nick into struct */
    safestrncpy(user->nick, new_nick, sizeof(user->nick));

    /* reinsert link according to new nick */
    dict_insert(clients, user->nick, user);

    if (IsLocal(user) && !no_announce) irc_nick(user);
}

struct userNode *
GetUserN(const char *numeric) /* using numeric */
{
    struct userNode *un;
    struct server *s;
    int n, slen, ulen;

    switch (strlen(numeric)) {
    default:
        log(MAIN_LOG, LOG_WARNING, "GetUserN(%s): numeric too long!\n", numeric);
        return NULL;
    case 5: slen = 2; ulen = 3; break;
    case 4: slen = 1; ulen = 3; break;
    case 3: slen = 1; ulen = 2; break;
    case 2: case 1: case 0:
	log(MAIN_LOG, LOG_WARNING, "GetUserN(%s): numeric too short!\n", numeric);
	return NULL;
    }
    if (!(s = servers_num[base64toint(numeric, slen)])) {
	log(MAIN_LOG, LOG_WARNING, "GetUserN(%s): couldn't find server (len=%d)!\n", numeric, slen);
	return NULL;
    }
    n = base64toint(numeric+slen, ulen) & s->num_mask;
    if (!(un = s->users[n])) {
	log(MAIN_LOG, LOG_WARNING, "GetUserN(%s) couldn't find user!\n", numeric);
    }
    return un;
}

struct userNode *
GetUserH(const char *nick)
{
    return dict_find(clients, nick, NULL);
}

struct userNode *
GetUser(const char *user)
{
    return (user[0] == ':') ? GetUserH(user+1) : GetUserN(user);
}

static new_channel_func_t *ncf_list;
static unsigned int ncf_size = 0, ncf_used = 0;

void
reg_new_channel_func(new_channel_func_t handler)
{
    if (ncf_used == ncf_size) {
	if (ncf_size) {
	    ncf_size <<= 1;
	    ncf_list = realloc(ncf_list, ncf_size*sizeof(ncf_list[0]));
	} else {
	    ncf_size = 8;
	    ncf_list = malloc(ncf_size*sizeof(ncf_list[0]));
	}
    }
    ncf_list[ncf_used++] = handler;
}

static join_func_t *jf_list;
static unsigned int jf_size = 0, jf_used = 0;

void
reg_join_func(join_func_t handler)
{
    if (jf_used == jf_size) {
	if (jf_size) {
	    jf_size <<= 1;
	    jf_list = realloc(jf_list, jf_size*sizeof(join_func_t));
	} else {
	    jf_size = 8;
	    jf_list = malloc(jf_size*sizeof(join_func_t));
	}
    }
    jf_list[jf_used++] = handler;
}

struct chanNode *
AddChannel(const char *name, time_t time_, const char *modes, const char *key, int limit, char *members, char *banlist)
{
    struct chanNode *cNode;
    int rel_age;
    unsigned int nn;

    if (!IsChannelName(name)) {
        log(MAIN_LOG, LOG_ERROR, "Somebody asked to add channel '%s', which isn't a channel name!\n", name);
        return NULL;
    }

    if (!(cNode = GetChannel(name))) {
        cNode = calloc(1, sizeof(*cNode));
        safestrncpy(cNode->name, name, sizeof(cNode->name));
        banList_init(&cNode->banlist);
        modeList_init(&cNode->members);
        dict_insert(channels, cNode->name, cNode);
    }

    if ((cNode->timestamp == 0) || (cNode->timestamp > time_)) {
        struct userList reops;
        unsigned int orig_limit;
        chan_mode_t orig_modes;
        char orig_key[KEYLEN+1];

        /* remember the old modes, and update them with the new */
        orig_modes = cNode->modes;
        orig_limit = cNode->limit;
        strcpy(orig_key, cNode->key);
        cNode->modes = 0;
        mod_chanmode(cNode, modes, key, limit);
        cNode->timestamp = time_;
        rel_age = 1;

        /* remove our old ban list, replace it with the new one */
        for (nn=0; nn<cNode->banlist.used; nn++) {
            free(cNode->banlist.list[nn]);
        }
        cNode->banlist.used = 0;

        /* deop anybody in the channel now, but re-op services */
        userList_init(&reops);
        for (nn=0; nn<cNode->members.used; nn++) {
            struct modeNode *mn = cNode->members.list[nn];
            if (mn->modes & MODE_CHANOP) {
                if (IsService(mn->user) && IsLocal(mn->user)) userList_append(&reops, mn->user);
                mn->modes &= ~MODE_CHANOP;
            }
        }
        if (reops.used) {
            nn = 0;
            /* Add ops back. */
            AddChannelOp(reops.used, reops.list, cNode, reops.list[0], 1);
            /* Turn on any mode bits that got turned off during the merge. */
            if ((orig_modes &= ~cNode->modes)) {
                char modebuf[MAXLEN];
                if (orig_modes & MODE_NOPRIVMSGS) modebuf[nn++] = 'n';
                if (orig_modes & MODE_TOPICLIMIT) modebuf[nn++] = 't';
                if (orig_modes & MODE_SECRET) modebuf[nn++] = 's';
                if (orig_modes & MODE_PRIVATE) modebuf[nn++] = 'p';
                if (orig_modes & MODE_INVITEONLY) modebuf[nn++] = 'i';
                if (orig_modes & MODE_MODERATED) modebuf[nn++] = 'm';
                if (orig_modes & MODE_DELAYJOINS) modebuf[nn++] = 'D';
                if (orig_modes & (MODE_LIMIT | MODE_KEY)) {
                    if (orig_modes & MODE_LIMIT) modebuf[nn++] = 'l';
                    if (orig_modes & MODE_KEY) modebuf[nn++] = 'k';
                    if (orig_modes & MODE_LIMIT) nn += sprintf(modebuf+nn, " %u", orig_limit);
                    if (orig_modes & MODE_KEY) nn += sprintf(modebuf+nn, " %s", orig_key);
                }
                modebuf[nn] = 0;
                mod_chanmode(cNode, modebuf, 0, 0);
                irc_mode(reops.list[0], cNode, modebuf);
            }
        }
        userList_clean(&reops);
    } else if (cNode->timestamp == time_) {
        rel_age = 0;
    } else {
        rel_age = -1;
    }

    /* rel_age is the relative ages of our channel data versus what is
     * in a BURST command.  1 means ours is younger, 0 means both are
     * the same age, -1 means ours is older. */

    /* if it's a new or updated channel, make callbacks */
    if (rel_age > 0) {
        for (nn=0; nn<ncf_used; nn++) {
            ncf_list[nn](cNode);
        }
    }

    /* go through list of bans and add each one */
    if (banlist && (rel_age >= 0)) {
        for (nn=0; banlist[nn];) {
            char *ban = banlist+nn;
            while (banlist[nn] != ' ' && banlist[nn]) nn++;
            while (banlist[nn] == ' ') banlist[nn++] = 0;
            AddChannelBan(1, &ban, cNode, NULL, now, 0);
        }
    }

    /* go through the list of users to add and add each one */
    if (members) {
        struct userNode *un;
        long mode;
        char *user, *end, sep;

        for (user = members, sep = *members, mode = 0; sep; user = end) {
            for (end = user + 3; (*end != ':') && (*end != ',') && *end; end++) ;
            sep = *end++; end[-1] = 0;
            if (sep == ':') {
                mode = 0;
                while ((sep = *end++)) {
                    if ((sep == 'o') && (rel_age >= 0)) mode |= MODE_CHANOP;
                    else if (sep == 'v') mode |= MODE_VOICE;
                    else break;
                }
            }
            if (!(un = GetUserN(user))) continue;
            AddChannelUser(un, cNode)->modes = mode;
        }
    }

    return cNode;
}

static del_channel_func_t *dcf_list;
static unsigned int dcf_size = 0, dcf_used = 0;

void
reg_del_channel_func(del_channel_func_t handler)
{
    if (dcf_used == dcf_size) {
	if (dcf_size) {
	    dcf_size <<= 1;
	    dcf_list = realloc(dcf_list, dcf_size*sizeof(dcf_list[0]));
	} else {
	    dcf_size = 8;
	    dcf_list = malloc(dcf_size*sizeof(dcf_list[0]));
	}
    }
    dcf_list[dcf_used++] = handler;
}

void
DelChannel(struct chanNode *channel)
{
    unsigned int n;
    char *argh;

    dict_remove(channels, channel->name);

    if (channel->members.used) {
        log(MAIN_LOG, LOG_ERROR, "Warning: deleting channel %s with %d users remaining\n", channel->name, channel->members.used);
    }

    /* go through all channel members and delete them from the channel */
    for (n=channel->members.used; n>0; ) {
	DelChannelUser(channel->members.list[--n]->user, channel, false, 1);
    }

    /* delete all channel bans */
    for (n=channel->banlist.used; n>0; ) {
	argh = channel->banlist.list[--n]->ban;
	DelChannelBan(1, &argh, channel, NULL, 0);
    }

    for (n=0; n<dcf_used; n++) {
        dcf_list[n](channel);
    }

    modeList_clean(&channel->members);
    banList_clean(&channel->banlist);
    free(channel);
}

struct modeNode *
AddChannelUser(struct userNode *user, struct chanNode* channel)
{
	struct modeNode *mNode;
	unsigned int n;

	mNode = GetUserMode(channel, user);
	if (mNode) return mNode; /* already in channel */

	mNode = malloc(sizeof(*mNode));

	/* set up modeNode */
	mNode->channel = channel;
	mNode->user = user;
	mNode->modes = 0;
        mNode->idle_since = now;

	/* add modeNode to channel and to user */
	modeList_append(&channel->members, mNode);
	modeList_append(&user->channels, mNode);

	if (IsLocal(user)) irc_join(user, channel);

        for (n=0; n<jf_used; n++) {
            jf_list[n](user, channel);
        }

	return mNode;
}

static part_func_t *pf_list;
static unsigned int pf_size = 0, pf_used = 0;

void
reg_part_func(part_func_t handler)
{
    if (pf_used == pf_size) {
	if (pf_size) {
	    pf_size <<= 1;
	    pf_list = realloc(pf_list, pf_size*sizeof(part_func_t));
	} else {
	    pf_size = 8;
	    pf_list = malloc(pf_size*sizeof(part_func_t));
	}
    }
    pf_list[pf_used++] = handler;
}

void
unreg_part_func(part_func_t handler)
{
    unsigned int i;
    for (i=0; i<pf_used; i++) {
        if (pf_list[i] == handler) break;
    }
    if (i == pf_used) return;
    memmove(pf_list+i, pf_list+i+1, (pf_used-i-1)*sizeof(pf_list[0]));
    pf_used--;
}

void
DelChannelUser(struct userNode* user, struct chanNode* channel, const char *reason, int deleting)
{
    struct modeNode* mNode;
    unsigned int n;

    if (reason) irc_part(user, channel, reason);

    mNode = GetUserMode(channel, user);

    /* Sometimes we get a PART when the user has been KICKed.
     * In this case, we get no usermode, and should not try to free it.
     */
    if (!mNode) return;

    /* remove modeNode from channel and user */
    modeList_remove(&channel->members, mNode);
    modeList_remove(&user->channels, mNode);
    free(mNode);

    for (n=0; n<pf_used; n++) pf_list[n](user, channel);

    if (!deleting && channel->members.used == 0) DelChannel(channel);
}

void
KickChannelUser(struct userNode* target, struct chanNode* channel, struct userNode *kicker, const char *why)
{
    if (!target || !channel || IsService(target) || !GetUserMode(channel, target)) return;
    /* don't remove them from the channel, since the server will send a PART */
    irc_kick(kicker, target, channel, why);

    if(IsLocal(target))
    {
	/* NULL reason because we don't want a PART message to be
	   sent by DelChannelUser. */
	DelChannelUser(target, channel, NULL, 0);
    }
}

static kick_func_t *kf_list;
static unsigned int kf_size = 0, kf_used = 0;

void
reg_kick_func(kick_func_t handler)
{
    if (kf_used == kf_size) {
	if (kf_size) {
	    kf_size <<= 1;
	    kf_list = realloc(kf_list, kf_size*sizeof(kick_func_t));
	} else {
	    kf_size = 8;
	    kf_list = malloc(kf_size*sizeof(kick_func_t));
	}
    }
    kf_list[kf_used++] = handler;
}

void
ChannelUserKicked(struct userNode* kicker, struct userNode* victim, struct chanNode* channel)
{
    unsigned int n;
    struct modeNode *mn;

    if (!kicker || !victim || !channel || IsService(victim) || !GetUserMode(channel, victim)) return;

    /* Update the kicker's idle time */
    if ((mn = GetUserMode(channel, kicker))) {
        mn->idle_since = now;
    }

    for (n=0; n<kf_used; n++) {
	kf_list[n](kicker, victim, channel);
    }

    DelChannelUser(victim, channel, 0, 0);

    if (IsLocal(victim)) {
	irc_part(victim, channel, NULL);
    }
}

static void
chan_apply_bans(struct userNode *who, struct chanNode *channel, char *what, unsigned int count, char **list)
{
    char buffer[400];
    unsigned int start, i, here_len, queued;

    queued = 0;
    start = sizeof(buffer)-1;
    buffer[start] = 0;
    for (i=0; i<count; i++) {
	here_len = strlen(list[i]);
	if ((start < queued+here_len+3) || (queued >= MAXMODEPARAMS)) {
	    while (queued) {
		buffer[--start] = what[1];
		queued--;
	    }
	    buffer[--start] = what[0];
	    irc_mode(who, channel, buffer+start);
	    start = sizeof(buffer)-1;
	}
	start -= here_len;
	memcpy(buffer+start, list[i], here_len); /* omit terminator */
	buffer[--start] = ' ';
	queued++;
    }
    if (queued) {
	while (queued) {
	    buffer[--start] = what[1];
	    queued--;
	}
	buffer[--start] = what[0];
	irc_mode(who, channel, buffer+start);
	start = sizeof(buffer)-1;
    }
}

static void
chan_apply_mode(struct userNode *who, struct chanNode *channel, char *what, unsigned int count, struct userNode **list)
{
    char buffer[400];
    unsigned int start, i, queued, len;

    queued = 0;
    start = sizeof(buffer)-1;
    buffer[start] = 0;

    for (i=0; i<count; i++) {
	if ((start < queued+3+COMBO_NUMERIC_LEN) || (queued >= MAXMODEPARAMS)) {
	    while (queued) {
		buffer[--start] = what[1];
		queued--;
	    }
	    buffer[--start] = what[0];
	    irc_mode(who, channel, buffer+start);
	    start = sizeof(buffer)-1;
	}
        len = strlen(list[i]->numeric);
	start -= len;
	memcpy(buffer+start, list[i]->numeric, len);
	buffer[--start] = ' ';
	queued++;
    }
    if (queued) {
	while (queued) {
	    buffer[--start] = what[1];
	    queued--;
	}
	buffer[--start] = what[0];
	irc_mode(who, channel, buffer+start);
	start = sizeof(buffer)-1;
    }
}

int ChannelBanExists(struct chanNode *channel, char *ban)
{
    unsigned int n;

    for(n = 0; n < channel->banlist.used; n++)
    {
	if(match_ircglobs(channel->banlist.list[n]->ban, ban))
	{
	    return 1;
	}
    }
    return 0;
}

void
AddChannelBan(unsigned int num, char **banset, struct chanNode* channel, struct userNode *who, time_t set, int announce)
{
    unsigned int n, i;
    struct banNode *bNode;
    struct string_list *remove = NULL;
    struct banNode **bans;

    for (n=0; n<num; n++) {
	if (channel->banlist.used >= MAXBANS) break;

	/* We check for bans that are a superset (or match) of an
	   existing ban here and silently remove the old ban to be
	   compatible with ircu; however, we _do_not_ check for bans
	   that are a subset of an existing ban here.  Instead, that
	   is checked in the user interface code (because it is
	   illegal to try to apply it). */
        bans = channel->banlist.list;
        for (i = 0; i < channel->banlist.used; i++) {
            if (match_ircglobs(banset[n], bans[i]->ban)) {
                if (!remove) remove = alloc_string_list(2);
                string_list_append(remove, strdup(bans[i]->ban));
            }
        }

	bNode = malloc(sizeof(struct banNode));
	
	safestrncpy(bNode->ban, banset[n], sizeof(bNode->ban));
	safestrncpy(bNode->who, (who ? who->nick : "<unknown>"), sizeof(bNode->who));
	bNode->set = set;
	banList_append(&channel->banlist, bNode);
    }

    /* If we have to announce the new ban, announce the
       removal of the old; if we aren't announcing the new
       ban (we received the ban from our uplink), do it
       silently. */
    if (remove) {
	DelChannelBan(remove->used, remove->list, channel, who, announce);
	free_string_list(remove);
    }
    if (announce) chan_apply_bans(who, channel, "+b", n, banset);
}

void
DelChannelBan(unsigned int num, char **banset, struct chanNode* channel, struct userNode *who, int announce)
{
    unsigned int n, m;

    if (announce) chan_apply_bans(who, channel, "-b", num, banset);
    for (n=0; n<num; n++) {
	for (m=0; m<channel->banlist.used; m++) {
	    struct banNode *bNode = channel->banlist.list[m];
	    if (!strcmp(bNode->ban, banset[n])) {
                banList_remove(&channel->banlist, bNode);
                free(bNode);
                break;
            }
	}
    }
}

void
AddChannelOp(unsigned int num, struct userNode **newops, struct chanNode *channel, struct userNode *who, int announce)
{
    unsigned int n;
    struct modeNode *mn;

    /* We need to announce before we op people so that we use the
     * right source numeric if a service is op'ing itself. */
    if (announce) chan_apply_mode(who, channel, "+o", num, newops);
    for (n=0; n<num; n++) {
	if (!(mn = GetUserMode(channel, newops[n]))) continue;
	mn->modes |= MODE_CHANOP;
    }
}

void
DelChannelOp(unsigned int num, struct userNode **exops, struct chanNode *channel, struct userNode *who, int announce)
{
    unsigned int n;
    struct modeNode *mn;

    /* We need to announce before we deop people so that we use the
     * right source numeric if a service is deop'ing itself. */
    if (announce) chan_apply_mode(who, channel, "-o", num, exops);
    for (n=0; n<num; n++) {
	if (!(mn = GetUserMode(channel, exops[n]))) continue;
	mn->modes &= ~MODE_CHANOP;
    }
}

void
AddChannelVoice(unsigned int num, struct userNode **newv, struct chanNode *channel, struct userNode *who, int announce)
{
    unsigned int n;
    struct modeNode *mn;

    for (n=0; n<num; n++) {
	if (!(mn = GetUserMode(channel, newv[n]))) continue;
	mn->modes |= MODE_VOICE;
    }
    if (announce) chan_apply_mode(who, channel, "+v", num, newv);
}

void
DelChannelVoice(unsigned int num, struct userNode **exv, struct chanNode *channel, struct userNode *who, int announce)
{
    unsigned int n;
    struct modeNode *mn;

    for (n=0; n<num; n++) {
	if (!(mn = GetUserMode(channel, exv[n]))) continue;
	mn->modes &= ~MODE_VOICE;
    }
    if (announce) chan_apply_mode(who, channel, "-v", num, exv);
}

static topic_func_t *tf_list;
static unsigned int tf_size = 0, tf_used = 0;

void
reg_topic_func(topic_func_t handler)
{
    if (tf_used == tf_size) {
	if (tf_size) {
	    tf_size <<= 1;
	    tf_list = realloc(tf_list, tf_size*sizeof(topic_func_t));
	} else {
	    tf_size = 8;
	    tf_list = malloc(tf_size*sizeof(topic_func_t));
	}
    }
    tf_list[tf_used++] = handler;
}

void
SetChannelTopic(struct chanNode *channel, struct userNode *user, const char *topic, int announce)
{
    unsigned int n;
    struct modeNode *mn;
    char old_topic[TOPICLEN+1];

    safestrncpy(old_topic, channel->topic, sizeof(old_topic));
    safestrncpy(channel->topic, topic, sizeof(channel->topic));
    safestrncpy(channel->topic_nick, user->nick, sizeof(channel->topic_nick));
    channel->topic_time = now;

    /* Update the setter's idle time */
    if ((mn = GetUserMode(channel, user))) {
        mn->idle_since = now;
    }

    if (announce) {
	/* We don't really care if a local user messes with the topic,
         * so don't call the tf_list functions. */
	irc_topic(user, channel, topic);
    } else {
	for (n=0; n<tf_used; n++) {
	    if (tf_list[n](user, channel, old_topic)) break;
	}
    }
}

struct chanNode *
GetChannel(const char *name)
{
    return dict_find(channels, name, NULL);
}

struct modeNode *
GetUserMode(struct chanNode *channel, struct userNode *user)
{
    unsigned int n;
    struct modeNode *mn = NULL;
    if (channel->members.used < user->channels.used) {
	for (n=0; n<channel->members.used; n++) {
	    if (user == channel->members.list[n]->user) {
		mn = channel->members.list[n];
		break;
	    }
	}
    } else {
	for (n=0; n<user->channels.used; n++) {
	    if (channel == user->channels.list[n]->channel) {
		mn = user->channels.list[n];
		break;
	    }
	}
    }
    return mn;
}

DEFINE_LIST(userList, struct userNode*)
DEFINE_LIST(modeList, struct modeNode*)
DEFINE_LIST(banList, struct banNode*)
DEFINE_LIST(channelList, struct chanNode*)
DEFINE_LIST(serverList, struct server*)

static void
hash_cleanup(void)
{
    unsigned int nn;

    DelServer(self, 0, NULL);
    dict_delete(channels);
    dict_delete(clients);
    dict_delete(servers);
    userList_clean(&curr_opers);

    if (slf_list) free(slf_list);
    if (nuf_list) free(nuf_list);
    if (ncf2_list) free(ncf2_list);
    if (duf_list) free(duf_list);
    if (ncf_list) free(ncf_list);
    if (jf_list) free(jf_list);
    if (dcf_list) free(dcf_list);
    if (pf_list) free(pf_list);
    if (kf_list) free(kf_list);
    if (tf_list) free(tf_list);

    for (nn=0; nn<dead_users.used; nn++) free(dead_users.list[nn]);
    userList_clean(&dead_users);
}
