/* recdb.c - recursive/record database implementation
 * Copyright 2000 srvx Development Team
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: recdb.c,v 1.36 2002/05/06 23:33:12 entrope Exp $
 */

#include "config.h"
#include <ctype.h>
#include <setjmp.h>
#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "dict.h"
#include "recdb.h"
#include "log.h"
#include "common.h"

#if HAVE_GETC_UNLOCKED
#undef fgetc
#define fgetc(STR) getc_unlocked(STR)
#endif

#if HAVE_PUTC_UNLOCKED
#undef fputc
#define fputc(CHR,STR) putc_unlocked(CHR,STR)
#endif

#if HAVE_FEOF_UNLOCKED
#undef feof
#define feof(STR) feof_unlocked(STR)
#endif

/* file format (grammar in Backus-Naur Form):
 *
 * database := record*
 * record := qstring [ '=' ] ( qstring | object | stringlist ) ';'
 * qstring := '"' ( [^\\] | ('\\' [\\n]) )* '"'
 * object := '{' record* '}'
 * stringlist := '(' [ qstring [',' qstring]* ] ')'
 *
 */

/* when a database or object is read from disk, it is represented as
 * a dictionary object, keys are names (what's left of the '=') and
 * values are 'struct record_data's
 */

struct recdb_context {
    int line;
    int col;
};

typedef struct recdb_file {
    const char *source;
    union {
	FILE *f;
	unsigned char *s;
    } s;
    int is_file : 1;
    struct recdb_context ctx;
    jmp_buf env;
} RECDB;

#define EOL '\n'

#if 1
#define ABORT(recdb, code, ch) longjmp((recdb)->env, ((code) << 8) | (ch))
#else
static void
ABORT(RECDB *recdb, int code, unsigned char ch) {
    longjmp(recdb->env, code << 8 | ch);
}
#endif

enum fail_codes {
    UNTERMINATED_STRING,
    EXPECTED_OPEN_QUOTE,
    EXPECTED_OPEN_BRACE,
    EXPECTED_OPEN_PAREN,
    EXPECTED_COMMA,
    EXPECTED_START_RECORD_DATA,
    EXPECTED_SEMICOLON,
    EXPECTED_RECORD_DATA
};

static void parse_record_int(RECDB *recdb, char **pname, struct record_data **prd);

/* allocation functions */

#define alloc_record_data_int() malloc(sizeof(struct record_data))

struct record_data *
alloc_record_data_qstring(const char *string)
{
    struct record_data *rd;
    rd = alloc_record_data_int();
    SET_RECORD_QSTRING(rd, string);
    return rd;
}

struct record_data *
alloc_record_data_object(dict_t obj)
{
    struct record_data *rd;
    rd = alloc_record_data_int();
    SET_RECORD_OBJECT(rd, obj);
    return rd;
}

struct record_data *
alloc_record_data_string_list(struct string_list *slist)
{
    struct record_data *rd;
    rd = alloc_record_data_int();
    SET_RECORD_STRING_LIST(rd, slist);
    return rd;
}

struct string_list*
alloc_string_list(int size)
{
    struct string_list *slist;
    slist = malloc(sizeof(struct string_list));
    slist->used = 0;
    slist->size = size;
    slist->list = slist->size ? malloc(size*sizeof(char*)) : NULL;
    return slist;
}

dict_t
alloc_database(void)
{
    dict_t db = dict_new();
    dict_set_free_data(db, free_record_data);
    return db;
}

/* misc. operations */

void
string_list_append(struct string_list *slist, char *string)
{
    if (slist->used == slist->size) {
        if (slist->size) {
            slist->size <<= 1;
            slist->list = realloc(slist->list, slist->size*sizeof(char*));
        } else {
            slist->size = 4;
            slist->list = malloc(slist->size*sizeof(char*));
        }
    }
    slist->list[slist->used++] = string;
}

struct string_list *
string_list_copy(struct string_list *slist)
{
    struct string_list *new_list;
    unsigned int i;
    new_list = alloc_string_list(slist->size);
    new_list->used = slist->used;
    for (i=0; i<new_list->used; i++) {
	new_list->list[i] = strdup(slist->list[i]);
    }
    return new_list;
}

struct record_data*
database_get_path(dict_t db, const char *path)
{
    char *new_path = strdup(path), *orig_path = new_path;
    char *part;
    struct record_data *rd;

    for (part=new_path; *new_path; new_path++) {
	if (*new_path != '/') continue;
	*new_path = 0;

	rd = dict_find(db, part, NULL);
	if (!rd || rd->type != RECDB_OBJECT) {
	    free(orig_path);
	    return NULL;
	}

	db = rd->d.object;
	part = new_path+1;
    }

    rd = dict_find(db, part, NULL);
    free(orig_path);
    return rd;
}

void*
database_get_data(dict_t db, const char *path, enum recdb_type type)
{
    struct record_data *rd = database_get_path(db, path);
    return (rd && rd->type == type) ? rd->d.whatever : NULL;
}

/* free functions */

void
free_string_list(struct string_list *slist)
{
    unsigned int i;
    for (i=0; i<slist->used; i++) free(slist->list[i]);
    if (slist->list) free(slist->list);
    slist->used = slist->size = 0;
    free(slist);
}

void
free_record_data(void *rdata)
{
    struct record_data *r = rdata;
    switch (r->type) {
    case RECDB_INVALID: break;
    case RECDB_QSTRING: free(r->d.qstring); break;
    case RECDB_OBJECT: dict_delete(r->d.object); break;
    case RECDB_STRING_LIST: free_string_list(r->d.slist); break;
    }
    free(r);
}

/* parse functions */

static int
dbeof(RECDB *recdb)
{
    return (recdb->is_file) ? feof(recdb->s.f) : !*recdb->s.s;
}

static int
dbgetc(RECDB *recdb)
{
    int res;
    if (recdb->is_file) {
	res = fgetc(recdb->s.f);
    } else {
	res = dbeof(recdb) ? EOF : *recdb->s.s++;
    }
    if (res == EOL) recdb->ctx.line++, recdb->ctx.col=1;
    else if (res != EOF) recdb->ctx.col++;
    return res;
}

static void
dbungetc(int c, RECDB *recdb)
{
    if (recdb->is_file) {
	ungetc(c, recdb->s.f);
    } else {
	*(--recdb->s.s) = c;
    }
}

/* returns first non-whitespace, non-comment character (-1 for EOF found) */
int
parse_skip_ws(RECDB *recdb)
{
    int c, d, in_comment = 0;
    while (!dbeof(recdb)) {
	c = dbgetc(recdb);
	if (c == EOF) return EOF;
	if (isspace(c)) continue;
	if (c != '/') return c;
	if ((d = dbgetc(recdb)) == '*') {
	    /* C style comment, with slash star comment star slash */
	    in_comment = 1;
	    do {
		do {
		    c = dbgetc(recdb);
		} while (c != '*' && c != EOF);
		if ((c = dbgetc(recdb)) == '/') in_comment = 0;
	    } while (in_comment);
	} else if (d == '/') {
	    /* C++ style comment, with slash slash comment newline */
	    do {
		c = dbgetc(recdb);
	    } while (c != EOF && c != EOL);
	} else {
	    if (d != EOF) dbungetc(d, recdb);
	    return c;
	}
    }
    return -1;
}

char *
parse_qstring(RECDB *recdb)
{
    char *buff;
    int used=0, size=8, c;
    struct recdb_context start_ctx;
    if ((c = parse_skip_ws(recdb)) == EOF) return NULL;
    start_ctx = recdb->ctx;
    if (c != '"') ABORT(recdb, EXPECTED_OPEN_QUOTE, c);
    buff = malloc(size);
    while (!dbeof(recdb) && (c = dbgetc(recdb)) != '"') {
	if (c != '\\') {
	    buff[used++] = c;
	} else {
	    switch (c = dbgetc(recdb)) {
	    case 'n': buff[used++] = EOL; break;
	    default: buff[used++] = c; break;
	    }
	}
	if (used == size) {
	    size <<= 1;
	    buff = realloc(buff, size);
	}
    }
    if (c != '"' && dbeof(recdb)) {
	free(buff);
	recdb->ctx = start_ctx;
	ABORT(recdb, UNTERMINATED_STRING, EOF);
    }
    buff[used] = 0;
    return buff;
}

dict_t
parse_object(RECDB *recdb)
{
    dict_t obj;
    char *name;
    struct record_data *rd;
    int c;
    if ((c = parse_skip_ws(recdb)) == EOF) return NULL;
    if (c != '{') ABORT(recdb, EXPECTED_OPEN_BRACE, c);
    obj = alloc_object();
    dict_set_free_keys(obj, free);
    while (!dbeof(recdb)) {
	if ((c = parse_skip_ws(recdb)) == '}') break;
	if (c == EOF) break;
        dbungetc(c, recdb);
	parse_record_int(recdb, &name, &rd);
	dict_insert(obj, name, rd);
    }
    return obj;
}

struct string_list *
parse_string_list(RECDB *recdb)
{
    struct string_list *slist;
    int c;
    if ((c = parse_skip_ws(recdb)) == EOF) return NULL;
    if (c != '(') ABORT(recdb, EXPECTED_OPEN_PAREN, c);
    slist = alloc_string_list(4);
    while (true) {
	c = parse_skip_ws(recdb);
	if (c == EOF || c == ')') break;
	dbungetc(c, recdb);
	string_list_append(slist, parse_qstring(recdb));
	c = parse_skip_ws(recdb);
	if (c == EOF || c == ')') break;
	if (c != ',') ABORT(recdb, EXPECTED_COMMA, c);
    }
    return slist;
}

static void
parse_record_int(RECDB *recdb, char **pname, struct record_data **prd)
{
    int c;
    *pname = parse_qstring(recdb);
    c = parse_skip_ws(recdb);
    if (c == EOF) {
	if (!*pname) return;
	free(*pname);
	ABORT(recdb, EXPECTED_RECORD_DATA, EOF);
    }
    if (c == '=') c = parse_skip_ws(recdb);
    dbungetc(c, recdb);
    *prd = malloc(sizeof(**prd));
    switch (c) {
    case '"':
        /* Don't use SET_RECORD_QSTRING, since that does an extra strdup() of the string. */
        (*prd)->type = RECDB_QSTRING;
        (*prd)->d.qstring = parse_qstring(recdb);
        break;
    case '{': SET_RECORD_OBJECT(*prd, parse_object(recdb)); break;
    case '(': SET_RECORD_STRING_LIST(*prd, parse_string_list(recdb)); break;
    default: ABORT(recdb, EXPECTED_START_RECORD_DATA, c);
    }
    if ((c = parse_skip_ws(recdb)) != ';') ABORT(recdb, EXPECTED_SEMICOLON, c);
}

static dict_t
parse_database_int(RECDB *recdb)
{
    char *name;
    struct record_data *rd;
    dict_t db = alloc_database();
    dict_set_free_keys(db, free);
    while (!dbeof(recdb)) {
	parse_record_int(recdb, &name, &rd);
        if (name) dict_insert(db, name, rd);
    }
    return db;
}

const char *
failure_reason(int code)
{
    const char *reason;
    switch (code >> 8) {
    case UNTERMINATED_STRING: reason = "Unterminated string"; break;
    case EXPECTED_OPEN_QUOTE: reason = "Expected '\"'"; break;
    case EXPECTED_OPEN_BRACE: reason = "Expected '{'"; break;
    case EXPECTED_OPEN_PAREN: reason = "Expected '('"; break;
    case EXPECTED_COMMA: reason = "Expected ','"; break;
    case EXPECTED_START_RECORD_DATA: reason = "Expected start of some record data"; break;
    case EXPECTED_SEMICOLON: reason = "Expected ';'"; break;
    case EXPECTED_RECORD_DATA: reason = "Expected record data"; break;
    default: reason = "Unknown error";
    }
    if (code == -1) reason = "Premature end of file";
    return reason;
}

void
explain_failure(RECDB *recdb, int code)
{
    log(MAIN_LOG, LOG_ERROR, "%s (got '%c') at %s line %d column %d\n",
	failure_reason(code), code & 255,
	recdb->source, recdb->ctx.line, recdb->ctx.col);
}

const char *
parse_record(const char *text, char **pname, struct record_data **prd)
{
    RECDB recdb;
    int res;
    *pname = NULL;
    *prd = NULL;
    recdb.source = "<user-supplied text>";
    recdb.s.s = strdup(text);
    recdb.is_file = 0;
    recdb.ctx.line = recdb.ctx.col = 1;
    if ((res = setjmp(recdb.env)) == 0) {
	parse_record_int(&recdb, pname, prd);
	return 0;
    } else {
	if (*pname) free(*pname);
	if (*prd) free(*prd);
	return failure_reason(res);
    }
}

dict_t
parse_database(const char *filename)
{
    RECDB recdb;
    int res;
    dict_t db;
    recdb.source = filename;
    recdb.s.f = fopen(filename, "r");
    if (!recdb.s.f) {
	log(MAIN_LOG, LOG_ERROR, "Unable to open database file '%s' for reading\n", filename);
	return NULL;
    }
    recdb.is_file = 1;
    recdb.ctx.line = recdb.ctx.col = 1;
    if ((res = setjmp(recdb.env)) == 0) {
	db = parse_database_int(&recdb);
    } else {
	explain_failure(&recdb, res);
        _exit(1);
    }
    fclose(recdb.s.f);
    return db;
}

/* output functions */

void
write_qstring(FILE *file, const char *str)
{
    fputc('"', file);
    while (*str) {
	switch (*str) {
	case EOL: fputs("\\n", file); break;
	case '\\': fputs("\\\\", file); break;
	case '"': fputs("\\\"", file); break;
	default: fputc(*str, file); break;
	}
	str++;
    }
    fputc('"', file);
}

static int
write_object_helper(const char *key, void *data, void *extra)
{
    write_record(extra, key, data);
    return 0;
}

void
write_object(FILE *file, dict_t obj)
{
    fputs("{ ", file);
    dict_foreach(obj, write_object_helper, file);
    fputs("}", file);
}

void
write_string_list(FILE *file, struct string_list *slist)
{
    unsigned int i;
    fputc('(', file);
    if (slist->used) {
	write_qstring(file, slist->list[0]);
	for (i=1; i<slist->used; i++) {
	    fputs(", ", file);
	    write_qstring(file, slist->list[i]);
	}
    }
    fputc(')', file);
}

void
write_record_data(FILE *file, struct record_data *rd)
{
    switch (rd->type) {
    case RECDB_INVALID: log(MAIN_LOG, LOG_ERROR, "tried to write untyped record data\n"); break;
    case RECDB_QSTRING: write_qstring(file, rd->d.qstring); break;
    case RECDB_OBJECT: write_object(file, rd->d.object); break;
    case RECDB_STRING_LIST: write_string_list(file, rd->d.slist); break;
    }
}

void
write_record(FILE *file, const char *name, struct record_data *rd)
{
    write_qstring(file, name);
    fputc(' ', file);
    write_record_data(file, rd);
    fputc(';', file);
    fputc((rd->type == RECDB_OBJECT) ? EOL : ' ', file);
}

static int
write_database_helper(const char *key, void *data, void *extra)
{
    write_record(extra, key, data);
    fputc(EOL, (FILE*)extra);
    return 0;
}

void
write_database(FILE *file, dict_t db)
{
    dict_foreach(db, write_database_helper, file);
}
